import itertools

import numpy as np
from h5py._hl.filters import guess_chunk
from ndindex import ChunkSize, Slice, Tuple
from numpy.testing import assert_equal
from pytest import mark, raises

from ..backend import (
    DEFAULT_CHUNK_SIZE,
    create_base_dataset,
    create_virtual_dataset,
    write_dataset,
    write_dataset_chunks,
)

CHUNK_SIZE_3D = 2**4  # = cbrt(DEFAULT_CHUNK_SIZE)


def test_initialize(setup_vfile):
    with setup_vfile() as f:
        pass
    f.close()


def test_create_base_dataset(h5file):
    create_base_dataset(h5file, "test_data", data=np.ones((DEFAULT_CHUNK_SIZE,)))
    assert h5file["_version_data/test_data/raw_data"].dtype == np.float64


def test_create_base_dataset_multidimension(h5file):
    create_base_dataset(
        h5file,
        "test_data",
        data=np.ones((CHUNK_SIZE_3D, CHUNK_SIZE_3D, 2)),
        chunks=(CHUNK_SIZE_3D, CHUNK_SIZE_3D, CHUNK_SIZE_3D),
    )
    assert h5file["_version_data/test_data/raw_data"].dtype == np.float64


def test_write_dataset(h5file):
    data1 = np.ones((2 * DEFAULT_CHUNK_SIZE,))
    data2 = np.concatenate(
        (
            2 * np.ones((DEFAULT_CHUNK_SIZE,)),
            2 * np.ones((DEFAULT_CHUNK_SIZE,)),
            3 * np.ones((DEFAULT_CHUNK_SIZE,)),
        )
    )
    slices1 = write_dataset(h5file, "test_data", data1)
    slices2 = write_dataset(h5file, "test_data", data2)

    # Chunk size is set by the size the first dataset
    chunksize = guess_chunk(data1.shape, None, data1.dtype.itemsize)[0]

    slices1_expected = {}
    for i in range(data1.size // chunksize):
        data_slice = (Slice(i * chunksize, (i + 1) * chunksize, 1),)
        slices1_expected[data_slice] = slice(0, chunksize)

    last_data1_idx = chunksize
    sorted_slices1 = sorted(slices1.items(), key=lambda x: x[0].raw[0].start)
    sorted_expected1 = sorted(slices1_expected.items(), key=lambda x: x[0][0].start)
    assert sorted_slices1 == sorted_expected1

    slices2_expected = {}

    for i in range(data2.size // chunksize):
        data_slice = (Slice(i * chunksize, (i + 1) * chunksize, 1),)

        if i * chunksize < 2 * DEFAULT_CHUNK_SIZE:
            # Handle first part of dataset
            slices2_expected[data_slice] = slice(
                last_data1_idx, last_data1_idx + chunksize
            )
        else:
            # Handle second part of dataset
            slices2_expected[data_slice] = slice(
                last_data1_idx + chunksize, last_data1_idx + 2 * chunksize
            )

    sorted_slices2 = sorted(slices2.items(), key=lambda x: x[0].raw[0].start)
    sorted_expected2 = sorted(slices2_expected.items(), key=lambda x: x[0][0].start)

    assert sorted_slices2 == sorted_expected2

    ds = h5file["/_version_data/test_data/raw_data"]

    # This will change depending on whether data1.size and data2.size evenly divide
    # chunksize.
    assert ds.shape == (3 * chunksize,)
    assert_equal(ds[0 : 1 * chunksize], 1.0)
    assert_equal(ds[1 * chunksize : 2 * chunksize], 2.0)
    assert_equal(ds[2 * chunksize : 3 * chunksize], 3.0)
    assert_equal(ds[3 * chunksize : 4 * chunksize], 0.0)
    assert ds.dtype == np.float64


def test_write_dataset_multidimension(h5file):
    chunks = 3 * (CHUNK_SIZE_3D,)
    data = np.zeros((2 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D))
    slices1 = write_dataset(h5file, "test_data", data, chunks=chunks)
    data2 = data.copy()
    for n, (i, j, k) in enumerate(itertools.product([0, 1], repeat=3)):
        data2[
            i * CHUNK_SIZE_3D : (i + 1) * CHUNK_SIZE_3D,
            j * CHUNK_SIZE_3D : (j + 1) * CHUNK_SIZE_3D,
            k * CHUNK_SIZE_3D : (k + 1) * CHUNK_SIZE_3D,
        ] = n

    slices2 = write_dataset(h5file, "test_data", data2, chunks=chunks)

    assert slices1 == {
        (
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
        ): slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D),
        (
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
        ): slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D),
        (
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
        ): slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D),
        (
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
        ): slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D),
        (
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
        ): slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D),
        (
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
        ): slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D),
        (
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
        ): slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D),
        (
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
        ): slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D),
    }
    assert slices2 == {
        (
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
        ): slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D),
        (
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
        ): slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D),
        (
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
        ): slice(2 * CHUNK_SIZE_3D, 3 * CHUNK_SIZE_3D),
        (
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
        ): slice(3 * CHUNK_SIZE_3D, 4 * CHUNK_SIZE_3D),
        (
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
        ): slice(4 * CHUNK_SIZE_3D, 5 * CHUNK_SIZE_3D),
        (
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
        ): slice(5 * CHUNK_SIZE_3D, 6 * CHUNK_SIZE_3D),
        (
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
        ): slice(6 * CHUNK_SIZE_3D, 7 * CHUNK_SIZE_3D),
        (
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
        ): slice(7 * CHUNK_SIZE_3D, 8 * CHUNK_SIZE_3D),
    }

    ds = h5file["/_version_data/test_data/raw_data"]
    assert ds.shape == (8 * CHUNK_SIZE_3D, CHUNK_SIZE_3D, CHUNK_SIZE_3D)
    for n in range(8):
        assert_equal(ds[n * CHUNK_SIZE_3D : (n + 1) * CHUNK_SIZE_3D], n)
    assert ds.dtype == np.float64


def test_write_dataset_chunks(h5file):
    data = np.ones((2 * DEFAULT_CHUNK_SIZE,))
    slices1 = write_dataset(h5file, "test_data", data)
    chunksize = guess_chunk(data.shape, None, data.dtype.itemsize)[0]

    raw_slice1 = Tuple(Slice(0 * chunksize, 1 * chunksize, 1))
    slices2 = write_dataset_chunks(
        h5file,
        "test_data",
        {
            Tuple(Slice(0 * chunksize, 1 * chunksize, 1)): slices1[raw_slice1],
            Tuple(Slice(1 * chunksize, 2 * chunksize, 1)): 2 * np.ones((chunksize,)),
            Tuple(Slice(2 * chunksize, 3 * chunksize, 1)): 2 * np.ones((chunksize,)),
            Tuple(Slice(3 * chunksize, 4 * chunksize, 1)): 3 * np.ones((chunksize,)),
        },
    )

    slices1_expected = {}
    for i in range(data.size // chunksize):
        data_slice = Tuple(Slice(i * chunksize, (i + 1) * chunksize, 1))
        slices1_expected[data_slice] = slice(0, chunksize)

    assert slices1 == slices1_expected
    assert slices2 == {
        Tuple(Slice(0 * chunksize, 1 * chunksize, 1)): slice(
            0 * chunksize, 1 * chunksize
        ),
        Tuple(Slice(1 * chunksize, 2 * chunksize, 1)): slice(
            1 * chunksize, 2 * chunksize
        ),
        Tuple(Slice(2 * chunksize, 3 * chunksize, 1)): slice(
            1 * chunksize, 2 * chunksize
        ),
        Tuple(Slice(3 * chunksize, 4 * chunksize, 1)): slice(
            2 * chunksize, 3 * chunksize
        ),
    }

    ds = h5file["/_version_data/test_data/raw_data"]
    assert ds.shape == (3 * chunksize,)
    assert_equal(ds[0 * chunksize : 1 * chunksize], 1.0)
    assert_equal(ds[1 * chunksize : 2 * chunksize], 2.0)
    assert_equal(ds[2 * chunksize : 3 * chunksize], 3.0)
    assert_equal(ds[3 * chunksize : 4 * chunksize], 0.0)
    assert ds.dtype == np.float64


def test_write_dataset_chunks_multidimension(h5file):
    chunks = ChunkSize(3 * (CHUNK_SIZE_3D,))
    shape = (2 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D)
    data = np.zeros(shape)
    slices1 = write_dataset(h5file, "test_data", data, chunks=chunks)
    data_dict = {}
    for n, c in enumerate(chunks.indices(shape)):
        if n == 0:
            data_dict[c] = slices1[c]
        else:
            data_dict[c] = n * np.ones(chunks)

    slices1 = write_dataset(h5file, "test_data", data, chunks=chunks)
    slices2 = write_dataset_chunks(h5file, "test_data", data_dict)

    assert slices1 == {
        c: slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D) for c in chunks.indices(shape)
    }
    assert slices2 == {
        c: slice(i * CHUNK_SIZE_3D, (i + 1) * CHUNK_SIZE_3D)
        for i, c in enumerate(chunks.indices(shape))
    }

    ds = h5file["/_version_data/test_data/raw_data"]
    assert ds.shape == (8 * CHUNK_SIZE_3D, CHUNK_SIZE_3D, CHUNK_SIZE_3D)
    for n in range(8):
        assert_equal(ds[n * CHUNK_SIZE_3D : (n + 1) * CHUNK_SIZE_3D], n)
    assert ds.dtype == np.float64


def test_write_dataset_offset(h5file):
    data1 = np.ones((2 * DEFAULT_CHUNK_SIZE,))
    data2 = np.concatenate(
        (
            2 * np.ones((DEFAULT_CHUNK_SIZE,)),
            2 * np.ones((DEFAULT_CHUNK_SIZE,)),
            3 * np.ones((DEFAULT_CHUNK_SIZE - 2,)),
        )
    )
    slices1 = write_dataset(h5file, "test_data", data1)
    slices2 = write_dataset(h5file, "test_data", data2)

    chunksize = guess_chunk(data1.shape, None, data1.dtype.itemsize)[0]

    slices1_expected = {}
    for i in range(data1.size // chunksize):
        data_slice = (Slice(i * chunksize, (i + 1) * chunksize, 1),)
        slices1_expected[data_slice] = slice(0, chunksize)

    last_data1_idx = chunksize
    slices2_expected = {}
    for i in range(data2.size // chunksize):
        data_slice = (Slice(i * chunksize, (i + 1) * chunksize, 1),)

        if i * chunksize < 2 * DEFAULT_CHUNK_SIZE:
            slices2_expected[data_slice] = slice(
                last_data1_idx, last_data1_idx + chunksize
            )
        else:
            slices2_expected[data_slice] = slice(
                last_data1_idx + chunksize, last_data1_idx + 2 * chunksize
            )

    n_remaining = data2.size % chunksize
    data_slice = (Slice((data2.size // chunksize) * chunksize, data2.size, 1),)
    slices2_expected[data_slice] = slice(
        last_data1_idx + 2 * chunksize,
        last_data1_idx + 2 * chunksize + n_remaining,
    )

    sorted_slices1 = sorted(slices1.items(), key=lambda x: x[0].raw[0].start)
    sorted_expected1 = sorted(slices1_expected.items(), key=lambda x: x[0][0].start)
    sorted_slices2 = sorted(slices2.items(), key=lambda x: x[0].raw[0].start)
    sorted_expected2 = sorted(slices2_expected.items(), key=lambda x: x[0][0].start)

    assert sorted_slices1 == sorted_expected1
    assert sorted_slices2 == sorted_expected2

    ds = h5file["/_version_data/test_data/raw_data"]
    assert ds.shape == (4 * chunksize,)
    assert_equal(ds[0 * chunksize : 1 * chunksize], 1.0)
    assert_equal(ds[1 * chunksize : 2 * chunksize], 2.0)
    assert_equal(ds[2 * chunksize : 3 * chunksize], 3.0)
    assert_equal(ds[2 * chunksize : 3 * chunksize], 3.0)
    assert_equal(ds[3 * chunksize : 4 * chunksize - 2], 3.0)
    assert_equal(ds[4 * chunksize - 2 : 4 * chunksize], 0.0)


def test_write_dataset_offset_multidimension(h5file):
    chunks = ChunkSize(3 * (CHUNK_SIZE_3D,))
    shape = (2 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D)
    data = np.zeros(shape)
    slices1 = write_dataset(h5file, "test_data", data, chunks=chunks)
    shape2 = (2 * CHUNK_SIZE_3D - 2, 2 * CHUNK_SIZE_3D - 2, 2 * CHUNK_SIZE_3D - 2)
    data2 = np.empty(shape2)
    for n, c in enumerate(chunks.indices(shape)):
        data2[c.raw] = n

    slices2 = write_dataset(h5file, "test_data", data2, chunks=chunks)

    assert slices1 == {
        (
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
        ): slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D),
        (
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
        ): slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D),
        (
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
        ): slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D),
        (
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
        ): slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D),
        (
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
        ): slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D),
        (
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
        ): slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D),
        (
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
        ): slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D),
        (
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 1),
        ): slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D),
    }

    assert slices2 == {
        (
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
        ): slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D),
        (
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D - 2, 1),
        ): slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D),
        (
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D - 2, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
        ): slice(2 * CHUNK_SIZE_3D, 3 * CHUNK_SIZE_3D),
        (
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D - 2, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D - 2, 1),
        ): slice(3 * CHUNK_SIZE_3D, 4 * CHUNK_SIZE_3D),
        (
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D - 2, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
        ): slice(4 * CHUNK_SIZE_3D, 5 * CHUNK_SIZE_3D - 2),
        (
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D - 2, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D - 2, 1),
        ): slice(5 * CHUNK_SIZE_3D, 6 * CHUNK_SIZE_3D - 2),
        (
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D - 2, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D - 2, 1),
            Slice(0 * CHUNK_SIZE_3D, 1 * CHUNK_SIZE_3D, 1),
        ): slice(6 * CHUNK_SIZE_3D, 7 * CHUNK_SIZE_3D - 2),
        (
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D - 2, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D - 2, 1),
            Slice(1 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D - 2, 1),
        ): slice(7 * CHUNK_SIZE_3D, 8 * CHUNK_SIZE_3D - 2),
    }

    ds = h5file["/_version_data/test_data/raw_data"]
    assert ds.shape == (8 * CHUNK_SIZE_3D, CHUNK_SIZE_3D, CHUNK_SIZE_3D)
    for n, c in enumerate(chunks.indices(shape2)):
        a = np.zeros(chunks)
        a[Tuple(*[slice(0, i) for i in shape2]).as_subindex(c).raw] = n
        assert_equal(ds[n * CHUNK_SIZE_3D : (n + 1) * CHUNK_SIZE_3D], a)
    assert ds.dtype == np.float64


@mark.setup_args(version_name="test_version")
def test_create_virtual_dataset(h5file):
    """Check that creating a virtual dataset from chunks of real datasets works."""
    data1 = np.ones((2 * DEFAULT_CHUNK_SIZE,))
    data2 = np.concatenate(
        (
            2 * np.ones((DEFAULT_CHUNK_SIZE,)),
            3 * np.ones((DEFAULT_CHUNK_SIZE,)),
        )
    )

    # Chunk size is set by the size the first dataset
    chunksize = guess_chunk(data1.shape, None, data1.dtype.itemsize)[0]

    with h5file as f:
        slices1 = write_dataset(f, "test_data", data1)
        slices2 = write_dataset(f, "test_data", data2)

        nchunks1 = int(np.ceil(2 * DEFAULT_CHUNK_SIZE / chunksize))

        # The virtual dataset contains all the data from slices1, and one chunk of data
        # from slices2
        virtual_data = create_virtual_dataset(
            f,
            "test_version",
            "test_data",
            ((nchunks1 + 1) * chunksize,),
            {
                **slices1,
                Tuple(
                    Slice(nchunks1 * chunksize, (nchunks1 + 1) * chunksize, 1),
                ): slices2[(Slice(1 * chunksize, 2 * chunksize, 1),)],
            },
        )

        assert virtual_data.shape == ((nchunks1 + 1) * chunksize,)
        assert_equal(virtual_data[0 : 2 * DEFAULT_CHUNK_SIZE], 1.0)
        assert_equal(virtual_data[2 * DEFAULT_CHUNK_SIZE : 3 * DEFAULT_CHUNK_SIZE], 2.0)
        assert virtual_data.dtype == np.float64


@mark.setup_args(version_name="test_version")
def test_create_virtual_dataset_attrs(h5file):
    data1 = np.ones((2 * DEFAULT_CHUNK_SIZE,))
    data2 = np.concatenate(
        (
            2 * np.ones((DEFAULT_CHUNK_SIZE,)),
            3 * np.ones((DEFAULT_CHUNK_SIZE,)),
        )
    )

    # Chunk size is set by the size the first dataset
    chunksize = guess_chunk(data1.shape, None, data1.dtype.itemsize)[0]

    with h5file as f:
        slices1 = write_dataset(f, "test_data", data1)
        slices2 = write_dataset(f, "test_data", data2)

        nchunks1 = int(np.ceil(2 * DEFAULT_CHUNK_SIZE / chunksize))

        attrs = {"attribute": "value"}
        # The virtual dataset contains all the data from slices1, and one chunk of data
        # from slices2
        virtual_data = create_virtual_dataset(
            f,
            "test_version",
            "test_data",
            ((nchunks1 + 1) * chunksize,),
            {
                **slices1,
                Tuple(
                    Slice(nchunks1 * chunksize, (nchunks1 + 1) * chunksize, 1),
                ): slices2[(Slice(1 * chunksize, 2 * chunksize, 1),)],
            },
            attrs=attrs,
        )

        assert dict(virtual_data.attrs) == {
            **attrs,
            "raw_data": "/_version_data/test_data/raw_data",
            "chunks": np.array([chunksize]),
        }


@mark.setup_args(version_name=["test_version1", "test_version2"])
def test_create_virtual_dataset_multidimension(h5file):
    chunks = 3 * (CHUNK_SIZE_3D,)
    shape = (2 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D)
    data = np.ones(shape)
    slices1 = write_dataset(h5file, "test_data", data, chunks=chunks)

    virtual_data = create_virtual_dataset(
        h5file, "test_version1", "test_data", shape, slices1
    )

    assert virtual_data.shape == shape
    assert_equal(virtual_data[:], 1)
    assert virtual_data.dtype == np.float64

    data2 = data.copy()
    for n, (i, j, k) in enumerate(itertools.product([0, 1], repeat=3)):
        data2[
            i * CHUNK_SIZE_3D : (i + 1) * CHUNK_SIZE_3D,
            j * CHUNK_SIZE_3D : (j + 1) * CHUNK_SIZE_3D,
            k * CHUNK_SIZE_3D : (k + 1) * CHUNK_SIZE_3D,
        ] = n

    slices2 = write_dataset(h5file, "test_data", data2, chunks=chunks)

    virtual_data2 = create_virtual_dataset(
        h5file, "test_version2", "test_data", shape, slices2
    )

    assert virtual_data2.shape == shape
    for n, (i, j, k) in enumerate(itertools.product([0, 1], repeat=3)):
        assert_equal(
            virtual_data2[
                i * CHUNK_SIZE_3D : (i + 1) * CHUNK_SIZE_3D,
                j * CHUNK_SIZE_3D : (j + 1) * CHUNK_SIZE_3D,
                k * CHUNK_SIZE_3D : (k + 1) * CHUNK_SIZE_3D,
            ],
            n,
        )
    assert virtual_data2.dtype == np.float64


@mark.setup_args(version_name="test_version")
def test_create_virtual_dataset_offset(h5file):
    data1 = np.ones((2 * DEFAULT_CHUNK_SIZE,))
    data2 = np.concatenate(
        (
            2 * np.ones((DEFAULT_CHUNK_SIZE,)),
            3 * np.ones((DEFAULT_CHUNK_SIZE - 2,)),
        )
    )

    slices1 = write_dataset(h5file, "test_data", data1)
    slices2 = write_dataset(h5file, "test_data", data2)

    # Chunk size is set by the size the first dataset
    chunksize = guess_chunk(data1.shape, None, data1.dtype.itemsize)[0]
    nchunks1 = int(np.ceil(data1.size / chunksize))
    nchunks2 = int(np.ceil(data2.size / chunksize))

    # After writing the data above, there is now 4 chunks in the raw dataset:
    #   raw_data[0*chunksize:1*chunksize] == 1.0
    #   raw_data[1*chunksize:2*chunksize] == 2.0
    #   raw_data[2*chunksize:3*chunksize] == 3.0
    #   raw_data[3*chunksize:4*chunksize-2] == 3.0
    # Create a virtual dataset including all data from the first dataset
    # and the last chunk of data from the second dataset.
    virtual_data = create_virtual_dataset(
        h5file,
        "test_version",
        "test_data",
        ((nchunks1 + 1) * chunksize - 2,),
        {
            **slices1,
            Tuple(
                Slice(nchunks1 * chunksize, (nchunks1 + 1) * chunksize - 2, 1),
            ): slices2[
                (Slice((nchunks2 - 1) * chunksize, nchunks2 * chunksize - 2, 1),)
            ],
        },
    )

    assert virtual_data.shape == ((nchunks1 + 1) * chunksize - 2,)
    assert_equal(virtual_data[0 : nchunks1 * chunksize], 1.0)
    assert_equal(
        virtual_data[nchunks1 * chunksize : (nchunks1 + 1) * chunksize - 2], 3.0
    )


@mark.setup_args(version_name="test_version")
def test_create_virtual_dataset_offset_multidimension(h5file):
    chunks = ChunkSize(3 * (CHUNK_SIZE_3D,))
    shape = (2 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D, 2 * CHUNK_SIZE_3D)
    data = np.zeros(shape)
    write_dataset(h5file, "test_data", data, chunks=chunks)
    shape2 = (2 * CHUNK_SIZE_3D - 2, 2 * CHUNK_SIZE_3D - 2, 2 * CHUNK_SIZE_3D - 2)
    data2 = np.empty(shape2)
    for n, c in enumerate(chunks.indices(shape)):
        data2[c.raw] = n

    slices2 = write_dataset(h5file, "test_data", data2, chunks=chunks)

    virtual_data = create_virtual_dataset(
        h5file, "test_version", "test_data", shape2, slices2
    )

    assert virtual_data.shape == shape2
    assert_equal(virtual_data[()], data2)
    assert virtual_data.dtype == np.float64


def test_write_dataset_chunk_size(h5file):
    chunk_size = 2**10
    chunks = (chunk_size,)
    slices1 = write_dataset(
        h5file, "test_data", np.ones((2 * chunk_size,)), chunks=chunks
    )
    raises(
        ValueError,
        lambda: write_dataset(h5file, "test_data", np.ones(chunks), chunks=(2**9,)),
    )
    slices2 = write_dataset_chunks(
        h5file,
        "test_data",
        {
            Tuple(Slice(0 * chunk_size, 1 * chunk_size, 1)): slices1[
                Tuple(Slice(0 * chunk_size, 1 * chunk_size, 1))
            ],
            Tuple(Slice(1 * chunk_size, 2 * chunk_size, 1)): 2 * np.ones((chunk_size,)),
            Tuple(Slice(2 * chunk_size, 3 * chunk_size, 1)): 2 * np.ones((chunk_size,)),
            Tuple(Slice(3 * chunk_size, 4 * chunk_size, 1)): 3 * np.ones((chunk_size,)),
        },
    )

    assert slices1 == {
        Tuple(Slice(0 * chunk_size, 1 * chunk_size, 1)): slice(
            0 * chunk_size, 1 * chunk_size
        ),
        Tuple(Slice(1 * chunk_size, 2 * chunk_size, 1)): slice(
            0 * chunk_size, 1 * chunk_size
        ),
    }
    assert slices2 == {
        Tuple(Slice(0 * chunk_size, 1 * chunk_size, 1)): slice(
            0 * chunk_size, 1 * chunk_size
        ),
        Tuple(Slice(1 * chunk_size, 2 * chunk_size, 1)): slice(
            1 * chunk_size, 2 * chunk_size
        ),
        Tuple(Slice(2 * chunk_size, 3 * chunk_size, 1)): slice(
            1 * chunk_size, 2 * chunk_size
        ),
        Tuple(Slice(3 * chunk_size, 4 * chunk_size, 1)): slice(
            2 * chunk_size, 3 * chunk_size
        ),
    }

    ds = h5file["/_version_data/test_data/raw_data"]
    assert ds.shape == (3 * chunk_size,)
    assert_equal(ds[0 : 1 * chunk_size], 1.0)
    assert_equal(ds[1 * chunk_size : 2 * chunk_size], 2.0)
    assert_equal(ds[2 * chunk_size : 3 * chunk_size], 3.0)
    assert_equal(ds[3 * chunk_size : 4 * chunk_size], 0.0)
    assert ds.dtype == np.float64


def test_write_dataset_offset_chunk_size(h5file):
    chunk_size = 2**10
    chunks = (chunk_size,)
    slices1 = write_dataset(
        h5file, "test_data", 1 * np.ones((2 * chunk_size,)), chunks=chunks
    )
    slices2 = write_dataset(
        h5file,
        "test_data",
        np.concatenate(
            (2 * np.ones(chunks), 2 * np.ones(chunks), 3 * np.ones((chunk_size - 2,)))
        ),
    )

    assert slices1 == {
        Tuple(Slice(0 * chunk_size, 1 * chunk_size, 1)): slice(
            0 * chunk_size, 1 * chunk_size
        ),
        Tuple(Slice(1 * chunk_size, 2 * chunk_size, 1)): slice(
            0 * chunk_size, 1 * chunk_size
        ),
    }
    assert slices2 == {
        Tuple(Slice(0 * chunk_size, 1 * chunk_size, 1)): slice(
            1 * chunk_size, 2 * chunk_size
        ),
        Tuple(Slice(1 * chunk_size, 2 * chunk_size, 1)): slice(
            1 * chunk_size, 2 * chunk_size
        ),
        Tuple(Slice(2 * chunk_size, 3 * chunk_size - 2, 1)): slice(
            2 * chunk_size, 3 * chunk_size - 2
        ),
    }

    ds = h5file["/_version_data/test_data/raw_data"]
    assert ds.shape == (3 * chunk_size,)
    assert_equal(ds[0 * chunk_size : 1 * chunk_size], 1.0)
    assert_equal(ds[1 * chunk_size : 2 * chunk_size], 2.0)
    assert_equal(ds[2 * chunk_size : 3 * chunk_size - 2], 3.0)
    assert_equal(ds[3 * chunk_size - 2 : 4 * chunk_size], 0.0)


def test_write_dataset_compression(h5file):
    data = np.ones((2 * DEFAULT_CHUNK_SIZE,))

    # Chunk size is set by the size the first dataset
    chunksize = guess_chunk(data.shape, None, data.dtype.itemsize)[0]
    nchunks = int(np.ceil(data.size / chunksize))

    slices1 = write_dataset(
        h5file, "test_data", data, compression="gzip", compression_opts=3
    )

    with raises(ValueError):
        write_dataset(
            h5file, "test_data", np.ones((DEFAULT_CHUNK_SIZE,)), compression="lzf"
        )

    with raises(ValueError):
        write_dataset(
            h5file,
            "test_data",
            np.ones((DEFAULT_CHUNK_SIZE,)),
            compression="gzip",
            compression_opts=4,
        )

    expected = {}
    for i in range(nchunks):
        expected[(Slice(i * chunksize, (i + 1) * chunksize, 1),)] = slice(0, chunksize)

    assert slices1 == expected
    ds = h5file["/_version_data/test_data/raw_data"]
    assert ds.shape == (chunksize,)
    assert_equal(ds[0:chunksize], 1.0)
    assert ds.dtype == np.float64
    assert ds.compression == "gzip"
    assert ds.compression_opts == 3


def test_create_empty_virtual_dataset(setup_vfile):
    """Check that creating an empty virtual dataset writes no raw data.

    Also check that the empty virtual dataset is formed correctly.
    See https://github.com/deshaw/versioned-hdf5/issues/314 for context.
    """
    name = "empty_dataset"

    with setup_vfile(version_name="r0") as f:
        write_dataset(f, "empty_dataset", np.array([]))
        create_virtual_dataset(
            f,
            "r0",
            name,
            (0,),
            {},
        )

        # Check that the raw data has only fill_value in it
        assert_equal(f["_version_data"][name]["raw_data"][:], 0.0)

        # Check that the virtual data is empty
        ds = f["_version_data"]["versions"]["r0"][name][:]
        assert_equal(ds, np.array([]))
        assert ds.shape == (0,)
        assert ds.size == 0



import hypothesis
import numpy as np
import numpy.ma as ma
from hypothesis import given
from hypothesis import strategies as st
from hypothesis.extra import numpy as stnp
from numpy.testing import assert_array_equal

from ..tools import asarray, ix_with_slices


def test_asarray():
    a = np.array([1, -1], dtype="i2")
    b = asarray(a)
    assert b is a
    b = asarray(a, "i2")
    assert b is a
    b = asarray(a, "u2")
    assert_array_equal(b, a.astype("u2"), strict=True)
    assert b.base is a
    b = asarray(a, "i4")
    assert_array_equal(b, a.astype("i4"), strict=True)
    assert b.base is None

    # Don't just test itemsize
    a = np.array([1, -1], dtype="i4")
    b = asarray(a, "f4")
    assert_array_equal(b, a.astype("f4"), strict=True)
    assert b.base is None

    # non-arrays are coerced to np.ndarray
    a = [1, -1]
    b = asarray(a)
    assert_array_equal(b, np.asarray(a), strict=True)
    b = asarray(a, dtype=np.float32)
    assert_array_equal(b, np.asarray(a, dtype=np.float32), strict=True)

    a = 1
    b = asarray(a)
    assert isinstance(b, np.ndarray)
    assert_array_equal(b, np.asarray(a), strict=True)

    a = np.int16(1)
    b = asarray(a)
    assert isinstance(b, np.ndarray)
    assert_array_equal(b, np.asarray(a), strict=True)

    # array-likes aren't coerced to np.ndarray
    a = ma.masked_array([1, -1], mask=[0, 1], dtype="i2")
    b = asarray(a)
    assert b is a
    b = asarray(a, dtype="u2")
    assert type(b) is type(a)
    assert b.base is a
    assert_array_equal(b, a.astype("u2"), strict=True)
    b = asarray(a, dtype="i4")
    assert type(b) is type(a)
    assert b.base is None
    assert_array_equal(b, a.astype("i4"), strict=True)


@st.composite
def ix_idx_st(draw, max_ndim: int = 4) -> tuple[tuple[int, ...], tuple]:
    """Hypothesis draw of slice, integer array, and boolean array indices, with no
    limitation on the number of fancy indices
    """

    def slices(size: int):
        start = st.one_of(st.none(), st.integers(-size - 1, size + 1))
        stop = st.one_of(st.none(), st.integers(-size - 1, size + 1))
        step = st.one_of(
            st.none(), st.integers(1, size + 1), st.integers(-size - 1, -1)
        )
        return st.builds(slice, start, stop, step)

    def fancy(size: int):
        return stnp.arrays(
            np.intp,
            shape=st.integers(0, size + 1),
            elements=st.integers(-size, size - 1),
            unique=False,
        )

    def mask(size: int):
        return stnp.arrays(np.bool_, shape=size, elements=st.booleans())

    shape_st = st.lists(st.integers(1, 10), min_size=1, max_size=max_ndim)
    shape = tuple(draw(shape_st))

    idx_st = st.tuples(
        *(st.one_of(slices(size), fancy(size), mask(size)) for size in shape)
    )
    return shape, draw(idx_st)


@given(ix_idx_st())
@hypothesis.settings(max_examples=1000)
def test_ix_with_slices(args):
    shape, idx = args
    a = np.arange(np.prod(shape)).reshape(shape)
    expect = a
    for axis, i in enumerate(idx):
        expect = expect[(slice(None),) * axis + (i,)]

    actual = a[ix_with_slices(*idx, shape=shape)]
    assert_array_equal(actual, expect)



from __future__ import absolute_import, division, print_function, with_statement

import datetime
import logging
import random
import sys
import time

import h5py
import numpy as np

sys.path.append("..")

from generate_data_base import TestDatasetPerformanceBase

from utils import temp_dir_ctx
from versioned_hdf5.api import VersionedHDF5File


class TestVersionedDatasetPerformance(TestDatasetPerformanceBase):
    @classmethod
    def _write_transactions_sparse(
        cls,
        name,
        chunk_size,
        compression,
        versions,
        print_transactions,
        num_rows_initial,
        num_transactions,
        num_rows_per_append,
        pct_changes,
        num_changes,
        pct_deletes,
        num_deletes,
        pct_inserts,
        num_inserts,
    ):
        logger = logging.getLogger(__name__)

        # with temp_dir_ctx() as tmp_dir:
        # filename = tmp_dir + f'/{name}.h5'
        filename = f"{name}.h5"
        tts = []
        f = h5py.File(filename, "w")
        told = time.time()
        t0 = told
        times = []
        try:
            if versions:
                file = VersionedHDF5File(f)
                with file.stage_version("initial_version") as group:
                    key0_ds = group.create_dataset(
                        name + "/key0",
                        data=np.random.rand(num_rows_initial),
                        dtype=(np.dtype("int64")),
                        chunks=chunk_size,
                        compression=compression,
                    )
                    key1_ds = group.create_dataset(
                        name + "/key1",
                        data=np.random.rand(num_rows_initial),
                        dtype=(np.dtype("int64")),
                        chunks=chunk_size,
                        compression=compression,
                    )
                    val_ds = group.create_dataset(
                        name + "/val",
                        data=np.random.rand(num_rows_initial),
                        dtype=(np.dtype("float64")),
                        chunks=chunk_size,
                        compression=compression,
                    )
            else:
                key0_ds = f.create_dataset(
                    name + "/key0",
                    data=np.random.rand(num_rows_initial),
                    dtype=(np.dtype("int64")),
                    maxshape=(None,),
                    chunks=(chunk_size,),
                    compression=compression,
                )
                key1_ds = f.create_dataset(
                    name + "/key1",
                    data=np.random.rand(num_rows_initial),
                    dtype=(np.dtype("int64")),
                    maxshape=(None,),
                    chunks=(chunk_size,),
                    compression=compression,
                )
                val_ds = f.create_dataset(
                    name + "/val",
                    data=np.random.rand(num_rows_initial),
                    dtype=(np.dtype("float64")),
                    maxshape=(None,),
                    chunks=(chunk_size,),
                    compression=compression,
                )

            for a in range(num_transactions):
                if print_transactions:
                    print("Transaction", a)
                tt = datetime.datetime.utcnow()
                if versions:
                    with file.stage_version(str(tt)) as group:
                        key0_ds = group[name + "/key0"]
                        key1_ds = group[name + "/key1"]
                        val_ds = group[name + "/val"]
                        cls._modify_dss_sparse(
                            key0_ds,
                            key1_ds,
                            val_ds,
                            num_rows_per_append,
                            pct_changes if a > 0 else 0.0,
                            num_changes,
                            pct_deletes if a > 0 else 0.0,
                            num_deletes,
                            pct_inserts if a > 0 else 0.0,
                            num_inserts,
                        )
                else:
                    cls._modify_dss_sparse(
                        key0_ds,
                        key1_ds,
                        val_ds,
                        num_rows_per_append,
                        pct_changes if a > 0 else 0.0,
                        num_changes,
                        pct_deletes if a > 0 else 0.0,
                        num_deletes,
                        pct_inserts if a > 0 else 0.0,
                        num_inserts,
                    )
                t = time.time()
                times.append(t - told)
                told = t
                tts.append(tt)
                logger.info("Wrote transaction %d at transaction time %s", a, tt)
                f.flush()
            times.append(t - t0)
        finally:
            f.close()
        return times

    @classmethod
    def _write_transactions_dense(
        cls,
        name,
        chunk_size,
        compression,
        versions,
        print_transactions,
        num_rows_initial_0,
        num_rows_initial_1,
        num_transactions,
        num_rows_per_append_0,
        pct_changes,
        num_changes,
        pct_deletes,
        num_deletes_0,
        num_deletes_1,
        pct_inserts,
        num_inserts_0,
        num_inserts_1,
    ):
        logger = logging.getLogger(__name__)

        # with temp_dir_ctx() as tmp_dir:
        # with f"/{name}.h5" as filename:
        filename = f"{name}.h5"
        tts = []
        f = h5py.File(filename, "w")
        told = time.time()
        t0 = told
        times = []
        try:
            if versions:
                file = VersionedHDF5File(f)
                with file.stage_version("initial_version") as group:
                    key0_ds = group.create_dataset(
                        name + "/key0",
                        data=np.random.rand(num_rows_initial_0),
                        dtype=(np.dtype("int64")),
                        chunks=chunk_size,
                        compression=compression,
                    )
                    key1_ds = group.create_dataset(
                        name + "/key1",
                        data=np.random.rand(num_rows_initial_1),
                        dtype=(np.dtype("int64")),
                        chunks=chunk_size,
                        compression=compression,
                    )
                    # two dimensional value array
                    val_ds = group.create_dataset(
                        name + "/val",
                        data=np.random.rand(num_rows_initial_0, num_rows_initial_1),
                        dtype=np.dtype("float64"),
                        chunks=(chunk_size, chunk_size),
                        compression=compression,
                    )
            else:
                key0_ds = f.create_dataset(
                    name + "/key0",
                    data=np.random.rand(num_rows_initial_0),
                    dtype=np.dtype("int64"),
                    maxshape=(None,),
                    chunks=(chunk_size,),
                    compression=compression,
                )
                key1_ds = f.create_dataset(
                    name + "/key1",
                    data=np.random.rand(num_rows_initial_0),
                    dtype=np.dtype("int64"),
                    maxshape=(None,),
                    chunks=(chunk_size,),
                    compression=compression,
                )
                val_ds = f.create_dataset(
                    name + "/val",
                    data=np.random.rand(num_rows_initial_0, num_rows_initial_1),
                    dtype=np.dtype("float64"),
                    maxshape=(None, None),
                    chunks=(chunk_size, chunk_size),
                    compression=compression,
                )

            for a in range(num_transactions):
                if print_transactions:
                    print(f"Transaction {a} of {num_transactions}")
                tt = datetime.datetime.utcnow()
                if versions:
                    with file.stage_version(str(tt)) as group:
                        key0_ds = group[name + "/key0"]
                        key1_ds = group[name + "/key1"]
                        val_ds = group[name + "/val"]
                        cls._modify_dss_dense(
                            key0_ds,
                            key1_ds,
                            val_ds,
                            num_rows_per_append_0,
                            pct_changes if a > 0 else 0.0,
                            num_changes,
                            pct_deletes if a > 0 else 0.0,
                            num_deletes_0,
                            num_deletes_1,
                            pct_inserts if a > 0 else 0.0,
                            num_inserts_0,
                            num_inserts_1,
                        )
                else:
                    cls._modify_dss_dense(
                        key0_ds,
                        key1_ds,
                        val_ds,
                        num_rows_per_append_0,
                        pct_changes if a > 0 else 0.0,
                        num_changes,
                        pct_deletes if a > 0 else 0.0,
                        num_deletes_0,
                        num_deletes_1,
                        pct_inserts if a > 0 else 0.0,
                        num_inserts_0,
                        num_inserts_1,
                    )

                t = time.time()
                times.append(t - told)
                told = t
                tts.append(tt)
                logger.info("Wrote transaction %d at transaction time %s", a, tt)
                f.flush()
            times.append(t - t0)
        finally:
            f.close()
        return times

    @classmethod
    def _write_transactions_dense_old(
        cls,
        name,
        num_rows_initial_0,
        num_rows_initial_1,
        num_transactions,
        num_rows_per_append_0,
        pct_changes,
        num_changes,
        pct_deletes,
        num_deletes_0,
        num_deletes_1,
        pct_inserts,
        num_inserts_0,
        num_inserts_1,
    ):
        logger = logging.getLogger(__name__)

        with temp_dir_ctx() as tmp_dir:
            filename = tmp_dir + f"/{name}.h5"
            tts = []
            f = h5py.File(filename, "w")
            file = VersionedHDF5File(f)
            try:
                with file.stage_version("initial_version") as group:
                    key0_ds = group.create_dataset(
                        name + "/key0",
                        data=np.random.rand(num_rows_initial_0),
                        dtype=(np.dtype("int64")),
                    )
                    key1_ds = group.create_dataset(
                        name + "/key1",
                        data=np.random.rand(num_rows_initial_1),
                        dtype=(np.dtype("int64")),
                    )
                    val_ds = group.create_dataset(
                        name + "/val",
                        data=np.random.rand(num_rows_initial_0 * num_rows_initial_1),
                        dtype=(np.dtype("float64")),
                    )
                for a in range(num_transactions):
                    tt = datetime.datetime.utcnow()
                    with file.stage_version(str(tt)) as group:
                        key0_ds = group[name + "/key0"]
                        key1_ds = group[name + "/key1"]
                        val_ds = group[name + "/val"]
                        cls._modify_dss_dense_old(
                            key0_ds,
                            key1_ds,
                            val_ds,
                            num_rows_per_append_0,
                            pct_changes if a > 0 else 0.0,
                            num_changes,
                            pct_deletes if a > 0 else 0.0,
                            num_deletes_0,
                            num_deletes_1,
                            pct_inserts if a > 0 else 0.0,
                            num_inserts_0,
                            num_inserts_1,
                        )

                    tts.append(tt)
                    logger.info("Wrote transaction %d at transaction time %s", a, tt)
            finally:
                f.close()

    @classmethod
    def _modify_dss_dense_old(
        cls,
        key0_ds,
        key1_ds,
        val_ds,
        num_rows_per_append_0,
        pct_changes,
        num_changes,
        pct_deletes,
        num_deletes_0,
        num_deletes_1,
        pct_inserts,
        num_inserts_0,
        num_inserts_1,
    ):
        n_key0 = len(key0_ds)
        n_key1 = len(key1_ds)
        n_val = len(val_ds)
        assert n_val == n_key0 * n_key1
        # change values
        if random.randrange(0, 100) <= pct_changes:
            r_num_chgs = int(np.random.randn() + num_changes)
            for b in range(r_num_chgs):
                r = random.randrange(0, n_val)
                val_ds[r] = np.random.rand()
        # delete rows
        if random.randrange(0, 100) <= pct_deletes:
            # delete from values in two steps

            # 1. delete from key0 and associated vals
            r_num_dels_0 = max(int(np.random.randn() + num_deletes_0), 1)
            rs_0 = [random.randrange(0, n_key0) for _ in range(r_num_dels_0)]

            rs_val = [r0 * n_key1 + r1 for r0 in rs_0 for r1 in range(n_key1)]
            n_val -= len(rs_val)
            arr_val = val_ds[:]
            arr_val = np.delete(arr_val, rs_val)

            n_key0 -= r_num_dels_0
            arr_key0 = key0_ds[:]
            arr_key0 = np.delete(arr_key0, rs_0)
            key0_ds.resize((n_key0,), refcheck=False)
            key0_ds[:] = arr_key0

            # 2. delete from key1 and associated vals
            r_num_dels_1 = max(int(np.random.randn() + num_deletes_1), 1)
            rs_1 = [random.randrange(0, n_key1) for _ in range(r_num_dels_1)]

            rs_val = [r0 * n_key1 + r1 for r0 in range(n_key0) for r1 in rs_1]
            n_val -= len(rs_val)
            arr_val = np.delete(arr_val, rs_val)
            val_ds.resize((n_val,), refcheck=False)
            val_ds[:] = arr_val

            n_key1 -= r_num_dels_1
            arr_key1 = key1_ds[:]
            arr_key1 = np.delete(arr_key1, rs_1)
            key1_ds.resize((n_key1,), refcheck=False)
            key1_ds[:] = arr_key1
        # insert rows
        if random.randrange(0, 100) <= pct_inserts:
            # insert into values in two steps
            arr_val = val_ds[:]

            # 1. insert into key0 and associated vals
            rand_num_inss_0 = max(int(np.random.randn() + num_inserts_0), 1)
            rs_0 = [random.randrange(0, n_key0) for _ in range(rand_num_inss_0)]

            rs_val = [r0 * n_key1 + r1 for r0 in rs_0 for r1 in range(n_key1)]
            n_val += len(rs_val)
            arr_val = np.insert(arr_val, rs_val, [np.random.rand() for _ in rs_val])

            arr_key0 = key0_ds[:]
            arr_key0 = np.insert(
                arr_key0, rs_0, np.random.randint(0, int(1e6), size=len(rs_0))
            )
            n_key0 += rand_num_inss_0
            key0_ds.resize((n_key0,), refcheck=False)
            key0_ds[:] = arr_key0

            # 2. insert into key1 and associated vals
            rand_num_inss_1 = max(int(np.random.randn() + num_inserts_1), 1)
            rs_1 = [random.randrange(0, n_key1) for _ in range(rand_num_inss_1)]

            rs_val = [r0 * n_key1 + r1 for r0 in range(n_key0) for r1 in rs_1]
            n_val += len(rs_val)
            arr_val = np.insert(arr_val, rs_val, np.random.rand(len(rs_val)))
            val_ds.resize((n_val,), refcheck=False)
            val_ds[:] = arr_val

            arr_key1 = key1_ds[:]
            arr_key1 = np.insert(
                arr_key1, rs_1, np.random.randint(0, int(1e6), size=len(rs_1))
            )
            n_key1 += rand_num_inss_1
            key1_ds.resize((n_key1,), refcheck=False)
            key1_ds[:] = arr_key1
        # append
        rand_num_apps_0 = int(np.random.randn() + num_rows_per_append_0)
        if rand_num_apps_0 > 0:
            # append to key0 and associated vals
            n_key0 += rand_num_apps_0
            key0_ds.resize((n_key0,), refcheck=False)
            key0_ds[-rand_num_apps_0:] = np.random.randint(
                0, int(1e6), size=rand_num_apps_0
            )

            num_val_apps = n_key1 * rand_num_apps_0
            n_val += num_val_apps
            val_ds.resize((n_val,), refcheck=False)
            val_ds[-num_val_apps:] = np.random.rand(num_val_apps)

    def test_mostly_appends_dense_old(
        self,
        num_transactions=250,
        filename="test_mostly_appends_dense_old",
        chunk_size=None,
        compression=None,
        versions=True,
        print_transactions=False,
        deterministic=False,
    ):
        num_rows_initial_0 = 30
        num_rows_initial_1 = 30
        num_rows_per_append_0 = 1

        if deterministic:
            pct_inserts = 0
            pct_deletes = 0
            pct_changes = 0
        else:
            pct_inserts = 5
            pct_deletes = 1
            pct_changes = 5

        num_inserts_0 = 1
        num_inserts_1 = 10
        num_deletes_0 = 1
        num_deletes_1 = 1
        num_changes = 10

        times = self._write_transactions_dense_old(
            filename,
            chunk_size,
            compression,
            versions,
            print_transactions,
            num_rows_initial_0,
            num_rows_initial_1,
            num_transactions,
            num_rows_per_append_0,
            pct_changes,
            num_changes,
            pct_deletes,
            num_deletes_0,
            num_deletes_1,
            pct_inserts,
            num_inserts_0,
            num_inserts_1,
        )
        return times


if __name__ == "__main__":
    num_transactions = [500]
    for t in num_transactions:
        times = TestVersionedDatasetPerformance().test_large_fraction_changes_sparse(
            t, print_transactions=True
        )
